/*!\file primitives.c
 *
 * \brief Primitives de dessin 2D sur texture GL
 *
 * A but pédagogique, ne gère qu'un seul écran.
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date March 14 2014
 */
#include "primitives.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/*!\brief pointeur vers les données texture-écran */
static Uint32 * _screen = NULL;
/*!\brief largeur et hauteur de la texture-écran */
static GLuint _screen_w = 2, _screen_h = 2, _tId = 0; 

/*!\brief La couleur en cours ; a modifier avec les fonctions setColor */
static Uint32 _cur_color = 0xFF0000FF;

/*!\brief revoie la \a _cur_color. */
extern Uint32 getCurColor(void) {
  return _cur_color;
}

/*!\brief met dans \a _cur_color la couleur passee en argument. */
extern void setColor32(Uint32 color) {
  _cur_color = color;
}

/*!brief initialise (ou réinitialise) l'écran aux dimensions \a w et
 * \a h, le lie à la texture tId */
extern void initScreen(GLuint tId, GLuint w, GLuint h) {
  assert(tId * w * h);
  if(_screen != NULL && (_screen_w != w || _screen_h != h)) {
    free(_screen);
    _screen = NULL;
  }
  _screen_w = w; _screen_h = h;
  glBindTexture(GL_TEXTURE_2D, _tId = tId);
  if(_screen == NULL) {
    _screen = calloc(_screen_w * _screen_h, sizeof *_screen);
    assert(_screen);
  }
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _screen_w, _screen_h, 0, GL_RGBA, GL_UNSIGNED_BYTE, _screen);
}

/*!\brief retourne la largeur de l'écran. */
extern GLuint getScreenWidth(void) {
  return _screen_w;
}

/*!\brief retourne la hauteur de l'écran. */
extern GLuint getScreenHeight(void) {
  return _screen_h;
}

/*!\brief Efface l'écran en y mettant la valeur \a color. */
extern void clearScreen(Uint32 color) {
  GLuint wh = _screen_w * _screen_h, i;
  for(i = 0; i < wh; i++)
    _screen[i] = color;
}

/*!brief met à jour l'écran en envoyant la sous texture de dimensions
 * \a rect à GL. Si rect vaut NULL, met à jour toute la texture.
 *
 * \todo gérer la sous-texture ; actuelement renvoie l'ensemble de la texture.
 */
extern void updateScreen(GLuint * rect) {
  glBindTexture(GL_TEXTURE_2D, _tId);
  if(rect == NULL)
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _screen_w, _screen_h, 0, GL_RGBA, GL_UNSIGNED_BYTE, _screen);
  else
    fprintf(stderr, "%s (%s:%d): à faire !\n", __func__, __FILE__, __LINE__);
}

/*!brief initialise l'écran aux dimensions \a w et \a h, le lie à la
 * texture tId */
extern void deleteScreen(void) {
  if(_screen != NULL) {
    free(_screen);
    _screen = NULL;
  }
}

/*!\brief renvoie la couleur à la coordonnée (x, y) */
extern Uint32 getPixel(int x, int y) {
  return _screen[y * _screen_w + x];
}

/*!\brief met la couleur en cours à la coordonnée (x, y) */
extern void putPixel(int x, int y) {
  _screen[y * _screen_w + x] = _cur_color;
}

/*!\brief dessine un segment (x0, y0) -> (x1, y1) de couleur \a
 * _cur_color en utilisant l'algorithme de Bresenham'65.
 */
extern void line(int x0, int y0, int x1, int y1) {
  int u = x1 - x0, v = y1 - y0;
  int pasX = (u < 0) ? -1 : 1, pasY = (v < 0) ? -1 : 1;
  int x, y, del, incH, incO;
  if(abs(u) < abs(v)) { /* deuxieme octant */
    del = (incH = ((pasX * u) << 1)) - pasY * v;
    incO = incH - ((pasY * v) << 1);
    for(y = y0, x = x0; y != y1; y += pasY) {
      if(IN_SCREEN(x, y))
	_screen[y * _screen_w + x] = _cur_color;
      if(del < 0) del += incH;
      else {
	del += incO;
	x += pasX;
      }
    }
  } else {  /* premier octant */
    del = (incH = ((pasY * v) << 1)) - pasX * u;
    incO = incH - ((pasX * u) << 1);
    for(x = x0, y = y0; x != x1; x += pasX) {
      if(IN_SCREEN(x, y))
	_screen[y * _screen_w + x] = _cur_color;
      if(del < 0) del += incH;
      else {
	del += incO;
	y += pasY;
      }
    }
  }
}

/*!\brief circle dessine un cercle centre en (x0, y0) de rayon r en
 * utilisant l'algorithme de Bresenham'77.
 */
extern void circle(int x0, int y0, int r) {
  int x, y, del, incH, incO, t = M_SQRT1_2 * r;
  del = 3 - (r << 1);
  incH = 6;
  incO = 10 - (r << 2);
  for(x = 0, y = r; x <= t; x++, incH += 4, incO += 4) {
    if(IN_SCREEN(x0 + x, y0 + y)) _screen[(y0 + y) * _screen_w + x0 + x] = _cur_color;
    if(IN_SCREEN(x0 + x, y0 - y)) _screen[(y0 - y) * _screen_w + x0 + x] = _cur_color;
    if(IN_SCREEN(x0 - x, y0 + y)) _screen[(y0 + y) * _screen_w + x0 - x] = _cur_color;
    if(IN_SCREEN(x0 - x, y0 - y)) _screen[(y0 - y) * _screen_w + x0 - x] = _cur_color;
    if(IN_SCREEN(x0 + y, y0 + x)) _screen[(y0 + x) * _screen_w + x0 + y] = _cur_color;
    if(IN_SCREEN(x0 + y, y0 - x)) _screen[(y0 - x) * _screen_w + x0 + y] = _cur_color;
    if(IN_SCREEN(x0 - y, y0 + x)) _screen[(y0 + x) * _screen_w + x0 - y] = _cur_color;
    if(IN_SCREEN(x0 - y, y0 - x)) _screen[(y0 - x) * _screen_w + x0 - y] = _cur_color;
    if(del < 0) del += incH;
    else{
      y--;
      incO += 4;
      del += incO;
    }
  }
}

/*!\brief dessine un segment horizontal (x0, y) -> (x1, y).
 */
extern void hLine(int x0, int x1, int y) {
  int x, pasX = (x1 - x0) < 0 ? -1 : 1;
  for(x = x0; x != x1; x += pasX)
    if(IN_SCREEN(x, y))	
      putPixel(x, y);
}

/*!\brief circle dessine un cercle plein centré en (x0, y0) de rayon r
 * en utilisant l'algorithme de Bresenham'77.
 */
extern void filledCircle(int x0, int y0, int r) {
  int x, y, del, incH, incO, t = M_SQRT1_2 * r;
  del = 3 - (r << 1);
  incH = 6;
  incO = 10 - (r << 2);
  for(x = 0, y = r; x <= t; x++, incH += 4, incO += 4) {
    hLine(x0 + x, x0 - x, y0 + y);
    hLine(x0 + x, x0 - x, y0 - y);
    hLine(x0 + y, x0 - y, y0 + x);
    hLine(x0 + y, x0 - y, y0 - x);
    if(del < 0) del += incH;
    else{
      y--;
      incO += 4;
      del += incO;
    }
  }
}
